from typing import Dict

import numpy
from typing_extensions import Final

RESOURCES_PATH: Final[str] = "appel/resources/"
DOTS_PER_CITY_TYPE = Dict[int, int]

DATASETS_TESTS_PATH: Final[str] = RESOURCES_PATH + "/test_points/"

TREE_PATH: Final[str] = "tree.bin"
CITIES_INFO_PATH: Final[str] = RESOURCES_PATH + "cidades_info_2016.csv.xz"
POLYGONS_PATHS: Final[Dict[str, str]] = {
    "STATES": RESOURCES_PATH + "estados",
    "CITIES": RESOURCES_PATH + "cidades",
    "MESOS": RESOURCES_PATH + "mesos",
    "MICROS": RESOURCES_PATH + "micros",
    "CITIES_ATTRS": RESOURCES_PATH + "cities_attrs.bin",
}
MAX_STATES: numpy.uint8 = numpy.uint8(27)
# Simplification values on each polygon

SIMPLIFICATION: float = 0.01



import unittest
from typing import List, Any, cast, Iterable, Union

import matplotlib.pyplot
import numpy
import shapely.geometry
from typing_extensions import Final

from appel.parser import Parser
from appel.region import Region


class ParserTestCase(unittest.TestCase):
    tree: numpy.ndarray

    def testGenerate(self) -> None:
        Parser.generate()
        self.testAll()

    def __plot(self) -> None:
        matplotlib.pyplot.plot(*self.tree[1].polygon.exterior.xy)
        matplotlib.pyplot.show()
        matplotlib.pyplot.plot(*self.tree[1][0].polygon.exterior.xy)
        matplotlib.pyplot.show()
        matplotlib.pyplot.plot(*self.tree[1][0][0].polygon.exterior.xy)
        matplotlib.pyplot.show()
        matplotlib.pyplot.plot(*self.tree[1][0][0][0].polygon.exterior.xy)
        matplotlib.pyplot.show()

    def __testcoordinate(self) -> None:
        xy: Final[Any] = self.tree[0][0][0][0].polygon.centroid.xy
        x = float(xy[0][0])
        y = float(xy[1][0])
        xl = -46.6481
        yl = -23.651
        x, y, xl, yl = [round(w, 3) for w in [x, y, xl, yl]]

        if x != xl or y != yl:
            raise RuntimeError("Coordenadas de sp erradas:", x, y, xl, yl)

    @staticmethod
    def __checkRegion(region: Region) -> None:
        if region.uid <= 0:
            raise RuntimeError("inválido" + region.__str__())
        if isinstance(region.polygon, shapely.geometry.MultiPolygon):
            for polygon in region.polygon:
                if not polygon.is_valid:
                    x: List[float] = []
                    y: List[float] = []
                    x += polygon.exterior.xy[0]
                    y += polygon.exterior.xy[1]
                    matplotlib.pyplot.plot(x, y)
                    matplotlib.pyplot.show()
        else:
            if not region.polygon.is_valid:
                print("region " + str(region.uid) + " not valid")
                x, y = region.polygon.xy
                matplotlib.pyplot.plot(x, y)
                matplotlib.pyplot.show()

    def __checkTree(self,
                    region: Union[Region, None] = None,
                    depth: int = 0) -> None:
        if region is None:
            for state in self.tree:
                self.__checkTree(state, 1)
        elif depth < 4:
            # noinspection Mypy
            for subregion in region:
                self.__checkRegion(subregion)
                self.__checkTree(subregion, depth + 1)
        else:
            self.__checkRegion(region)

    def __simplifiedPlot(self) -> None:
        city = self.tree[0][0]
        x: List[float] = []
        y: List[float] = []
        for region in city:
            x_, y_ = None, None
            if isinstance(region.polygon, shapely.geometry.MultiPolygon):
                for polygon in region.polygon:
                    x_, y_ = polygon.exterior.xy
            else:
                x_, y_ = region.polygon.exterior.xy
            x += list(cast(Iterable[float], x_))
            y += list(cast(Iterable[float], y_))

            matplotlib.pyplot.plot(x, y)
        matplotlib.pyplot.show()

    def testAll(self) -> None:
        self.tree = Parser.read()

        self.__plot()
        self.__testcoordinate()
        self.__checkTree()
        self.__simplifiedPlot()



from math import floor
from pickle import dump
from typing import List, Tuple

import geopandas
import numpy.random.mtrand
import pandas
import shapely
import shapely.geometry
import shapely.vectorized
from numpy import ndarray, array
from typing_extensions import Final

from appel import config, parser

TEST_RANGE: Final[ndarray] = array([10000, 100000, 1000000])
BRASIL_POPULATION: Final[int] = 206081432


def __decreaseToFitTestSize(test_size: int,
                            lats_final: List[float],
                            longs_final: List[float]) -> Tuple[List[float],
                                                               List[float]]:
    if len(lats_final) > test_size:
        return lats_final[0:test_size], longs_final[0:test_size]
    else:
        return lats_final, longs_final


def __checkConsistency(test_size: int,
                       lats_final: List[float],
                       longs_final: List[float],
                       p: shapely.geometry.Polygon) -> None:
    if len(lats_final) != len(longs_final):
        raise RuntimeError("Quantidade de longitudes e latitudes diferentes")

    matched_test: Final[ndarray] = shapely.vectorized.contains(p, longs_final, lats_final)

    if matched_test.sum() < test_size:
        raise RuntimeError("Ainda faltam alguns pontos", matched_test.sum(), test_size)


def randomDatasetGenerate() -> None:
    for test_size in TEST_RANGE:
        lats_final: List[float] = []
        longs_final: List[float] = []
        brp: shapely.geometry.Polygon = geopandas.read_file(config.RESOURCES_PATH + 'brasil').geometry.iloc[0]
        (minx, miny, maxx, maxy) = brp.bounds  # type: float, float, float, float

        while len(lats_final) < test_size:
            lats: ndarray = numpy.random.mtrand.uniform(miny, maxy, test_size)

            longs: ndarray = numpy.random.mtrand.uniform(minx, maxx, test_size)

            matched: ndarray = shapely.vectorized.contains(brp, longs, lats)

            lats_final += lats[matched].tolist()
            longs_final += longs[matched].tolist()

        lats_final, longs_final = __decreaseToFitTestSize(test_size, lats_final, longs_final)
        __checkConsistency(test_size, lats_final, longs_final, brp)

        with open('rand' + str(test_size) + '.bin', 'wb') as file:
            dump((longs_final, lats_final), file)


def gaussianDatasetGenerate() -> None:  # NOSONAR
    cities_info: pandas.DataFrame = \
        pandas.read_csv(config.CITIES_INFO_PATH, engine='c')[['Código do Município', 'População (Nº de habitantes)']]
    cities_info = cities_info.set_index('Código do Município')

    for test_size in TEST_RANGE:

        tree = parser.Parser.read()
        lats_final: List[float] = []
        longs_final: List[float] = []

        for state in tree:
            for meso in state:
                for micro in meso:
                    for city in micro:
                        polygon: shapely.geometry.Polygon = city.polygon
                        (minx, miny, maxx, maxy) = polygon.bounds

                        lats_local: List[float] = []
                        longs_local: List[float] = []
                        size_local = floor((cities_info.loc[city.uid] / BRASIL_POPULATION) * test_size)

                        while len(lats_local) < size_local:
                            lats = numpy.random.mtrand.uniform(miny, maxy, size_local)

                            longs = numpy.random.mtrand.uniform(minx, maxx, size_local)

                            conts = shapely.vectorized.contains(polygon, longs, lats)

                            lats_local += lats[conts].tolist()
                            longs_local += longs[conts].tolist()

                        lats_final, longs_final = __decreaseToFitTestSize(test_size, lats_final, longs_final)

                        __checkConsistency(test_size, lats_local, longs_local, polygon)

                        lats_final += lats_local
                        longs_final += longs_local

        with open('gauss' + str(test_size) + '.bin', 'wb') as file:
            dump((array(longs_final).astype('float16'), array(lats_final).astype('float16')), file)



from typing import Set, List, Union, TypeVar, cast

import numpy
import shapely.geometry
from typing_extensions import Final

TRegion = TypeVar('TRegion', bound='Region')


class Region(object):
    """A city representation. Will store just the needed information"""

    uid: Final[numpy.uint64]
    polygon: Union[shapely.geometry.Polygon, shapely.geometry.MultiPolygon]
    population: numpy.uint32
    upper: Final[numpy.uint64]  # Temporary attribute
    ids: Final[Set[int]]  # Temporary attribute
    # noinspection Mypy
    subregions: Final[List[Union[TRegion]]]  # Cities doesn't have this attribute

    def __init__(
            self,
            uid: numpy.uint64,
            upper: numpy.uint64,
            population: numpy.uint32,
            polygon: shapely.geometry.Polygon,

    ):
        """
        :param polygon: Shapefile polygon representing the Region
        """
        self.uid = uid
        self.population = population
        self.ids = set()
        self.subregions = []
        self.upper = upper
        self.polygon = polygon

    def __getitem__(self, item: int) -> TRegion:
        return self.subregions[item]

    def __str__(self) -> str:
        return str(self.uid + len(self.subregions) + self.upper)

    def __gt__(self, other: TRegion) -> bool:
        """Useful for numpy sort"""
        return cast(bool, self.population > other.population)

    def __lt__(self, other: TRegion) -> bool:
        """Useful for numpy sort"""
        return cast(bool, self.population < other.population)

    def add(self, subregion: TRegion) -> None:
        """
        Add subregion to dataset and sum it's population to our own
        :param subregion: Region
        :return:
        """
        if subregion.uid not in self.ids:
            self.population += subregion.population
            self.subregions.append(subregion)
            self.ids.add(subregion.uid)
        else:
            raise RuntimeError("Tried to add the same id: " + str(subregion.uid) + "to this region:" + str(type(self)))

    def clean(self, final_region: bool) -> None:
        """
        Delete unused memories after parsing
        """
        del self.ids
        if final_region:
            del self.subregions
        del self.population
        del self.upper

    def sort(self) -> None:
        """
        Sort region based on it's population
        """
        self.subregions.sort(reverse=True)

    def contains(self, point: shapely.geometry.Point) -> bool:
        """
        Check if a point is inside our polygon
        :param point: shapely Point
        :return: True if it's inside False if not
        """
        return cast(bool, self.polygon.contains(point))



import os
import pickle
import statistics
import time
import unittest
from typing import Dict, List, Callable, cast, Tuple, Any

import geopandas
import numpy
import pandas
import psycopg2.extras
import shapely.geometry
from typing_extensions import Final

from appel import config
from appel.searchtree import SearchTree


class CorrespondenceTestCase(unittest.TestCase):
    @staticmethod
    def __checkTestInputArray(xs: numpy.ndarray, ys: numpy.ndarray) -> None:
        assert isinstance(xs, numpy.ndarray)
        assert isinstance(ys, numpy.ndarray)
        assert len(xs) == len(ys)
        xs.astype('float16')
        ys.astype('float16')

    # noinspection SqlResolve,SqlNoDataSourceInspection
    def postGIS(self, xs: numpy.ndarray, ys: numpy.ndarray) -> Tuple[float, pandas.DataFrame]:
        """
        Batch method for quering the postgresql test_name database
        :return: Return the query results: List containing the row of info
        """
        self.__checkTestInputArray(xs, ys)

        connection = psycopg2.connect(
            "dbname='gabrielcoimbra' user='gabrielcoimbra' host='nesped' "
            "password='pass'"
        )

        cur = connection.cursor()
        xs, ys = xs.tolist(), ys.tolist()
        start: Final[float] = time.time()

        sql = "INSERT INTO test_points VALUES (0, ST_SetSRID(ST_MakePoint(%s, %s), 4674))"

        psycopg2.extras.execute_batch(cur, sql, zip(cast(List[float], xs), cast(List[float], ys)))

        cur.execute("CREATE INDEX sidx_points ON test_points USING GIST (geom);")

        cur.execute(
            """select simplified_geometries_cidades.uid 
            from simplified_geometries_cidades, test_points 
            where ST_Contains(simplified_geometries_cidades.geom, test_points.geom)""")
        stop: Final[float] = time.time()

        results: List[Any] = list(cur)

        cur.execute("TRUNCATE test_points")
        cur.execute("DROP INDEX sidx_points")
        connection.commit()
        cur.close()

        return stop - start, pandas.DataFrame(results)

    def geoPandasTest(self, xs: numpy.ndarray, ys: numpy.ndarray) -> Tuple[float, pandas.DataFrame]:
        self.__checkTestInputArray(xs, ys)

        pontos: List[shapely.geometry.Point] = []
        for i in range(len(xs)):
            pontos.append(shapely.geometry.Point(xs[i], ys[i]))
        cidades: Final = geopandas.GeoDataFrame(geopandas.read_file(config.POLYGONS_PATHS['CITIES'])
                                                [['geometry', 'uid']],
                                                crs={'init': 'epsg:4674'})

        points: Final = geopandas.GeoDataFrame(geometry=pontos, crs={'init': 'epsg:4674'})
        start: Final[float] = time.time()
        results: Final[pandas.DataFrame] = geopandas.sjoin(points, cidades, how="inner", op='within')
        return time.time() - start, results

    def appelTest(self, xs: numpy.ndarray, ys: numpy.ndarray) -> Tuple[float, pandas.DataFrame]:
        self.__checkTestInputArray(xs, ys)

        correspondence = SearchTree()
        start: Final[float] = time.time()
        results: Final[pandas.DataFrame] = correspondence.query(xs, ys)
        return time.time() - start, results

    def test(self) -> None:

        for filename in os.listdir(config.RESOURCES_PATH + '/test_points/'):
            if not filename.endswith('.bin'):
                continue
            with open(config.RESOURCES_PATH + '/test_points/' + filename, 'rb') as file:
                testcase = pickle.load(file)

            methods: List[Callable[[numpy.ndarray,
                                    numpy.ndarray], Tuple[float, pandas.DataFrame]]] = \
                [
                    self.appelTest,
                    # self.postGIS,
                    # self.geoPandasTest
                ]

            tempos_dic: Dict[str, List[float]] = {}
            for _ in range(4):
                for method in methods:
                    tempo: float = method(testcase[0], testcase[1])[0]
                    try:
                        tempos_dic[method.__name__].append(tempo)
                    except KeyError:
                        tempos_dic[method.__name__] = [tempo]

            for nome, tempos in tempos_dic.items():
                media, desvio = [str(round(x, ndigits=2)).replace('.', ',')
                                 for x in (statistics.mean(tempos), statistics.stdev(tempos))]
                print(nome, media, desvio)



import pickle
from typing import Dict, Union, Sized

import geopandas
import numpy
import pandas
from typing_extensions import Final

from appel import config, region


class Parser(object):
    """Stores a global access for a prefixed
    This is uddsed to ensure a
    """

    # array sorted by population containing tree
    tree = numpy.empty(config.MAX_STATES, dtype=object)

    @staticmethod
    def __mergeToUpperRegionAndSort(uppers: Dict[int, region.Region],
                                    lowers: Dict[int, region.Region]) -> None:
        def add(x: region.Region) -> None:
            try:
                uppers[x.upper].add(x)
            except KeyError as e:
                print("Não ache poly", e)
            except AttributeError as e:
                print("Erro de atributo", e)

        list(map(add, lowers.values()))

        # noinspection Mypy
        list(map(lambda x: x.sort(), uppers.values()))

    @classmethod
    def __readAndValidateAndSetTypesGDF(cls, path: str, upper_cast: bool = True,
                                        simplification: Union[float, None] = None) -> geopandas.GeoDataFrame:

        gdf: Final[geopandas.GeoDataFrame] = geopandas.read_file(path, engine="c")

        if simplification is None:
            gdf.geometry = gdf.geometry.simplify(config.SIMPLIFICATION)
        else:
            gdf.geometry = gdf.geometry.simplify(simplification)

        gdf.dropna(inplace=True)
        if upper_cast:
            # Tem que ser antes de passar para unsigned, pois checa valores negativos
            cls.__checkGDFValidity(gdf)
            gdf.uid = gdf.uid.astype('uint64', copy=False)
            gdf.upper = gdf.upper.astype('uint64', copy=False)

        print("Dropando\n ", gdf[gdf.isnull()])
        gdf.uid = gdf.uid.astype('uint64', copy=False)

        return gdf

    @staticmethod
    def __checkGDFValidity(gdf: geopandas.GeoDataFrame) -> None:
        """
        Auto explicatorio
        """
        inv: Final[Sized] = gdf[(gdf.uid.astype('int') <= 0) |
                                (gdf.upper.astype('int') <= 0)]
        if len(inv) > 0:
            print(inv)
            print("Dataset errado em uid ou upper.", len(inv))

        inv_poly: Final[pandas.DataFrame] = gdf[gdf.is_valid]
        if (len(gdf) - len(inv_poly)) > 0:
            print(inv_poly)
            print("Runtime polígonos inválidos", len(gdf) - len(inv_poly))

    @classmethod
    def _getCities(cls) -> Dict[int, region.Region]:
        """
        Get cities information and it's polygons
        :return: Dict[int, Region]
        """
        # xz loads faster
        cinfo: Final[pandas.DataFrame] = pandas.read_csv(config.CITIES_INFO_PATH, engine="c")

        cities: Dict[int, region.Region] = {}
        for _, row in cinfo.iterrows():
            uid: numpy.uintc = numpy.uint32(row["Código do Município"])
            population: numpy.uintc = numpy.uint32(
                str(row["População (Nº de habitantes)"]).replace(".", "")
            )
            cities[uid] = population

        gdf: Final[geopandas.GeoDataFrame] = cls.__readAndValidateAndSetTypesGDF(config.POLYGONS_PATHS["CITIES"],
                                                                                 simplification=0.005)
        for row in gdf.itertuples():
            cities[row.uid] = region.Region(row.uid, row.upper, numpy.uint32(cities[row.uid]), row.geometry)

        return cities

    @classmethod
    def _getMicros(cls) -> Dict[int, region.Region]:
        """
        Map microrregions
        :return: Dict[int, Region]
        """
        gdf: Final[geopandas.GeoDataFrame] = cls.__readAndValidateAndSetTypesGDF(config.POLYGONS_PATHS["MICROS"])

        micros: Dict[int, region.Region] = {}
        for row in gdf.itertuples():
            micros[row.uid] = region.Region(row.uid, row.upper, numpy.uint32(0), row.geometry)

        cls.__mergeToUpperRegionAndSort(micros, cls._getCities())

        return micros

    @classmethod
    def _getMesos(cls) -> Dict[int, region.Region]:
        """
        Map mesosregions
        :return: Dict[int, Region]
        """
        gdf: Final[geopandas.GeoDataFrame] = cls.__readAndValidateAndSetTypesGDF(config.POLYGONS_PATHS["MESOS"])

        mesos: Dict[int, region.Region] = {}
        for row in gdf.itertuples():
            mesos[row.uid] = region.Region(row.uid, row.upper, numpy.uint32(0), row.geometry)

        cls.__mergeToUpperRegionAndSort(mesos, cls._getMicros())

        return mesos

    @classmethod
    def _getStates(cls) -> Dict[int, region.Region]:
        """
        Iterate datasets
        Itera datasets para obter polígonos das cidades e somar população
        :return: Mapping id estado -> Estado
        """
        # Mapping id estado -> Estado

        gdf: Final[geopandas.GeoDataFrame] = cls.__readAndValidateAndSetTypesGDF(config.POLYGONS_PATHS["STATES"],
                                                                                 upper_cast=False)
        states: Dict[int, region.Region] = {}

        # Get tree polygon
        for row in gdf.itertuples():
            states[row.uid] = region.Region(row.uid, numpy.uint64(0), numpy.uint32(0), row.geometry)

        cls.__mergeToUpperRegionAndSort(states, cls._getMesos())

        return states

    @classmethod
    def _clean(cls, subregion: region.Region) -> None:
        if not subregion.subregions:
            subregion.clean(True)
        else:
            subregion.clean(False)
            # noinspection Mypy
            list(map(lambda x: cls._clean(x), subregion.subregions))

    @classmethod
    def generate(cls) -> None:
        """
        Generate the search structure for the cities.
        """
        # Pega dicionário não ordenado
        states: Final[Dict[int, region.Region]] = cls._getStates()

        cls.tree = numpy.array([x for x in states.values()])
        # Ordena os estados em si
        cls.tree[::-1].sort()

        # noinspection Mypy
        list(map(lambda x: cls._clean(x), cls.tree))

        # Escreve em disco
        cls._write()

    @classmethod
    def _write(cls) -> None:
        """ Write to disk generated tree list"""
        with open(config.TREE_PATH, "wb") as file:
            pickle.dump(cls.tree, file, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def read() -> numpy.ndarray:
        """ Read from disk generated tree"""
        with open(config.TREE_PATH, "rb") as file:
            return pickle.load(file)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="appel_geocode",
    version="0.1.1",
    author="Gabriel Coimbra",
    author_email="gcoimbra@riseup.net",
    description="A faster spatial join/reverse geocoding algorithm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gcoimbra/appel",
    packages=setuptools.find_packages(),
    install_requires=[
        'matplotlib==3.1.1',
        'Shapely==1.6.4.post2',
        'geopandas==0.5.1',
        'numpy==1.17.2',
        'pandas==0.25.1',
        'typing-extensions==3.7.4'

    ],

    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.6',
)



import math
import multiprocessing
import pickle
from typing import List, Union, Any

import numpy
import pandas
import shapely.vectorized
from typing_extensions import Final


class SearchTree:
    tree: numpy.ndarray
    THREAD_TRIGGER: Final[int] = 100
    MAX_WAIT: Final[float] = 200.0

    def __init__(self, tree_path: str = 'tree.bin') -> None:
        """
        Read state database generated tree list
        :return: array sorted by population containing tree
        """
        # Não precisa carregar se já estiver no disco
        with open(tree_path, "rb") as file:
            self.tree = pickle.load(file)

    def _wait(self, events: numpy.ndarray, tasks: List[multiprocessing.Process]) -> None:
        """
        Wait for threads to finish
        :return:
        """
        # _wait for all processes to finish

        for event in events:
            if not event.wait(timeout=self.MAX_WAIT):
                for task in tasks:
                    task.terminate()
                raise RuntimeError("Dataset demorou muito tempo para processar!")

    def query(self, longitudes: numpy.ndarray, latitudes: numpy.ndarray) -> pandas.DataFrame:
        """
        Multiprocessed function for looking up which city a point is on.
        divided in three phases map -> _wait -> reduce.
        :return: dictionary to   map: city uid -> test_points counted
        """

        ds_size: Final[int] = len(longitudes)
        if ds_size < self.THREAD_TRIGGER:
            return self._doMap(None, None, latitudes, longitudes, False)
        else:
            # noinspection Mypy
            queue: multiprocessing.Queue = multiprocessing.Queue()

            events: numpy.ndarray = numpy.empty(multiprocessing.cpu_count(), dtype=object)
            chunk_size: Final[int] = numpy.uint32(math.ceil(len(longitudes) / multiprocessing.cpu_count()))
            tasks: List[multiprocessing.Process] = []

            for i in range(multiprocessing.cpu_count()):
                events[i] = multiprocessing.Event()
                begin: int = chunk_size * i
                end: int = chunk_size * (i + 1)

                task: multiprocessing.Process = multiprocessing.Process(
                    target=self._doMap,
                    args=(
                        queue,
                        events[i],
                        longitudes[begin:end],
                        latitudes[begin:end],
                        True
                    ),
                )
                tasks.append(task)
                task.start()

            self._wait(events, tasks)
            return self._doReduce(queue)

    # noinspection Mypy
    @staticmethod
    def _doReduce(queue: multiprocessing.Queue) -> pandas.DataFrame:
        """
        Reduce phase: if different thread found a point on the same city
        this functions sum all these equals id
        This subfunction has access to the main function scope.
        :return:
        """
        results: List[pandas.DataFrame] = []
        for _ in range(queue.qsize()):
            results.append(queue.get())

        return pandas.concat(results)

    @staticmethod
    def __finishMap(xs: List[numpy.ndarray],
                    ys: List[numpy.ndarray],
                    cids: List[int]) -> pandas.DataFrame:

        return pandas.DataFrame({'uid': pandas.Series(cids, dtype='uint64'),
                                 'longitude': numpy.concatenate(xs).astype('float32', copy=False),
                                 'latitude': numpy.concatenate(ys).astype('float32', copy=False)})

    # Complexidade é necessária para questões de performance
    # noinspection Mypy
    def _doMap(  # NOSONAR
            self,
            queue: Union[multiprocessing.Queue, None],
            event: Union[Any, None],  # TODO: type callable how?
            x: numpy.ndarray,
            y: numpy.ndarray,
            parallel: bool
    ) -> Union[pandas.DataFrame, None]:

        """
        """
        # Initilize not found count
        ds_size: Final[int] = len(x)
        uids: List[int] = []
        lats: List[numpy.ndarray] = []
        longs: List[numpy.ndarray] = []
        states_acc: int = 0
        for state in self.tree:
            l0 = shapely.vectorized.contains(state.polygon, x, y)
            x0, y0 = x[l0], y[l0]
            states_found = x[l0].shape[0]
            states_acc += states_found
            if states_found > 0:
                meso_acc = 0
                for meso in state:
                    l1 = shapely.vectorized.contains(meso.polygon, x0, y0)
                    x01, y01 = x0[l1], y0[l1]
                    mesos_found = x01.shape[0]
                    meso_acc += mesos_found
                    if mesos_found > 0:
                        micro_acc = 0
                        for micro in meso:
                            l2 = shapely.vectorized.contains(
                                micro.polygon, x01, y01
                            )
                            x012, y012 = x01[l2], y01[l2]
                            micros_found = x012.shape[0]
                            micro_acc += micros_found
                            if micros_found > 0:
                                city_acc = 0
                                for city in micro:
                                    l3 = shapely.vectorized.contains(
                                        city.polygon, x012, y012
                                    )
                                    x0123, y0123 = x012[l3], y012[l3]
                                    city_acc += x0123.shape[0]
                                    if city_acc > 0:
                                        uids += [city.uid for x in range(len(x0123))]
                                        longs.append(x0123)
                                        lats.append(y0123)

                                    if city_acc == micros_found:
                                        break
                            if mesos_found == micro_acc:
                                break
                    if meso_acc == states_found:
                        break
            if states_acc == ds_size:
                break

        results: Final[pandas.DataFrame] = self.__finishMap(longs, lats, uids)
        if parallel:

            # noinspection Mypy
            queue.put(results)

            # noinspection Mypy
            event.set()
            return None
        else:
            return results
